/*
 * Copyright (c) 2016-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include <sys/time.h>
#include <math.h>

#include "mrsim_c_types.h"
#include "codels.h"

/* --- mrsim_motor_pwm ----------------------------------------------------- */

void
mrsim_motor_pwm(struct mrsim_ids_simstate_s_motor_s *motor, const double pwm)
{
  struct timeval tv;

  if (motor->disabled) return;
  if (!motor->spinning) return;
  gettimeofday(&tv, NULL);

  if (motor->servo) {
    motor->servo = false;
    motor->wd = 0;
    motor->smc_bias = 0;
    motor->smc_gain = 0;
    motor->smc_err = 0;
  }
  motor->pwm = pwm;
  motor->wdts.sec = tv.tv_sec;
  motor->wdts.nsec = 1000 * tv.tv_usec;
}


/* --- mrsim_motor_period -------------------------------------------------- */

void
mrsim_motor_period(struct mrsim_ids_simstate_s_motor_s *motor, const double wd)
{
  struct timeval tv;

  if (motor->disabled) return;
  if (!motor->spinning) return;
  gettimeofday(&tv, NULL);

  if (!motor->servo) {
    motor->servo = true;
    motor->smc_bias = motor->pwm;
    motor->smc_gain = 1;
    motor->smc_err = 0;
  }
  motor->wd = wd;
  motor->wdts.sec = tv.tv_sec;
  motor->wdts.nsec = 1000 * tv.tv_usec;
}


/* --- mrsim_motor_control ------------------------------------------------- */

void
mrsim_motor_control(const struct mrsim_ids_simparam_s_motor_s *p,
                    struct mrsim_ids_simstate_s_motor_s *motor)
{
  if (motor->starting) {
    /* startup PWM for 20Hz */
    motor->pwm =
      (2*M_PI * (p->K + p->R * p->b / p->K) / p->V) * 20;
    if (motor->w > 15.) {
      motor->spinning = true;
      motor->starting = false;
      mrsim_motor_period(motor, 1e6/65535);
    }
    return;
  }

  if (!motor->spinning) {
    motor->pwm = 0.;
    return;
  }

  if (motor->servo) {
    double w;
    int16_t u;
    uint16_t t, p;

    w = mrsim_motor_measure(motor);

    p = fabs(w) < 1000000./65535 ? 65535 : 1000000./fabs(w);
    p = p/42;

    t = fabs(motor->wd) < 1000000./65535 ? 65535 : 1000000./fabs(motor->wd);
    t = t/42;

    /* error sign */
    if (t < p)
      motor->smc_err = (3 * motor->smc_err - 65536) / 4;
    else if (t > p)
      motor->smc_err = (3 * motor->smc_err + 65536) / 4;
    else
      motor->smc_err = 3 * motor->smc_err / 4;

    /* gain adaptation */
    if (motor->smc_err < -50000 || motor->smc_err > 50000) {
      if (motor->smc_gain < 512*motor->pwm) motor->smc_gain += 4;
    } else if (motor->smc_err < -32768 || motor->smc_err > 32768) {
      if (motor->smc_gain < 512*motor->pwm) motor->smc_gain += 2;
    } else {
      motor->smc_gain -= 2;
      if (motor->smc_gain < 1) motor->smc_gain = 1;
    }

    /* bias adaptation */
    if (motor->smc_err < -65000) {
      motor->smc_bias += 6;
      if (motor->smc_bias > 1023) motor->smc_bias = 1023;
    } else if (motor->smc_err < -50000) {
      if (motor->smc_bias < 1023) motor->smc_bias += 1;
    } else if (motor->smc_err > 65000) {
      motor->smc_bias -= 6;
      if (motor->smc_bias < 1) motor->smc_bias = 1;
    } else if (motor->smc_err > 50000) {
      if (motor->smc_bias > 1) motor->smc_bias -= 1;
    }

    /* command */
    if (t < p)
      u = motor->smc_bias + motor->smc_gain;
    else if (t > p)
      u = motor->smc_bias - motor->smc_gain;
    else
      u = motor->smc_bias;

    if (u > 1023) u = 1023;
    else if (u < 0) u = 0;

    motor->pwm =  u/1024.;
    return;
  }
}


/* --- mrsim_motor_measure ------------------------------------------------- */

double
mrsim_motor_measure(const struct mrsim_ids_simstate_s_motor_s *motor)
{
  uint16_t p;

  p = fabs(motor->w) < 1000000./65535 ? 65535 : 1000000./fabs(motor->w);
  p = p/42 + gnoise(2);

  return 1000000./42/p;
}


/* --- mrsim_seqid --------------------------------------------------------- */

uint8_t
mrsim_seqid(struct timeval tv, double period)
{
  double t = tv.tv_sec - 1400000000; /* offset for better double precision */
  t += tv.tv_usec * 1e-6;

  return fmod(t/period, 256);
}
